from unittest.mock import Mock
import pytest
import warnings

from typing import Any, List, Optional

from requests.sessions import Session as Session

from posit.connect.resources import Resource, Resources

config = Mock()
session = Mock()


class FakeResource(Resource):
    @property
    def foo(self) -> Optional[str]:
        return self.get("foo")


class TestResource:
    def test_init(self):
        k = "foo"
        v = "bar"
        d = dict({k: v})
        r = FakeResource(config, session, **d)
        assert r.session == session
        assert r.config == config

    def test__getitem__(self):
        warnings.filterwarnings("ignore", category=FutureWarning)
        k = "foo"
        v = "bar"
        d = dict({k: v})
        r = FakeResource(config, session, **d)
        assert r.__getitem__(k) == d.__getitem__(k)
        assert r[k] == d[k]

    def test__setitem__(self):
        warnings.filterwarnings("ignore", category=FutureWarning)
        k = "foo"
        v1 = "bar"
        v2 = "baz"
        d = dict({k: v1})
        r = FakeResource(config, session, **d)
        assert r[k] == v1
        r[k] = v2
        assert r[k] == v2

    def test__delitem__(self):
        warnings.filterwarnings("ignore", category=FutureWarning)
        k = "foo"
        v = "bar"
        d = dict({k: v})
        r = FakeResource(config, session, **d)
        assert k in r
        assert r[k] == v
        del r[k]
        assert k not in r

    def test_foo(self):
        k = "foo"
        v = "bar"
        d = dict({k: v})
        r = FakeResource(config, session, **d)
        assert r.foo == v


class TestResources(Resources[Any]):
    def create(self) -> Any:
        return super().create()  # type: ignore [safe-super]

    def delete(self) -> None:
        return super().delete()  # type: ignore [safe-super]

    def find(self) -> List[Any]:
        return super().find()  # type: ignore [safe-super]

    def find_one(self) -> Optional[Any]:
        return super().find_one()  # type: ignore [safe-super]

    def get(self) -> Any:
        return super().get()  # type: ignore [safe-super]

    def update(self) -> Any:
        return super().update()  # type: ignore [safe-super]

    def count(self) -> int:
        return super().count()  # type: ignore [safe-super]

    def test_create(self):
        with pytest.raises(NotImplementedError):
            self.create()

    def test_delete(self):
        with pytest.raises(NotImplementedError):
            self.delete()

    def test_find(self):
        with pytest.raises(NotImplementedError):
            self.find()

    def test_find_one(self):
        with pytest.raises(NotImplementedError):
            self.find_one()

    def test_get(self):
        with pytest.raises(NotImplementedError):
            self.get()

    def test_update(self):
        with pytest.raises(NotImplementedError):
            self.update()

    def test_count(self):
        with pytest.raises(NotImplementedError):
            self.count()



import warnings

from abc import ABC, abstractmethod
from typing import Any, Generic, List, Optional, TypeVar

import requests

from .config import Config


T = TypeVar("T")


class Resource(ABC, dict):
    def __init__(self, config: Config, session: requests.Session, **kwargs):
        super().__init__(**kwargs)
        self.config: Config
        super().__setattr__("config", config)
        self.session: requests.Session
        super().__setattr__("session", session)

    def __getitem__(self, key):
        warnings.warn(
            f"__getitem__ for '{key}' does not support backwards compatibility. Consider using field based access instead: 'instance.{key}'",
            FutureWarning,
        )
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        warnings.warn(
            f"__setitem__ for '{key}' does not support backwards compatibility. Consider using field based access instead: 'instance.{key} = {value}",
            FutureWarning,
        )
        return super().__setitem__(key, value)

    def __delitem__(self, key):
        warnings.warn(
            f"__delitem__ for '{key}' does not support backwards compatibility. Consider using field based access instead: 'del instance.{key}'",
            FutureWarning,
        )
        return super().__delitem__(key)

    def __setattr__(self, name: str, value: Any) -> None:
        raise AttributeError("cannot set attributes: use update() instead")


class Resources(ABC, Generic[T]):
    @abstractmethod
    def create(self, *args, **kwargs) -> T:
        raise NotImplementedError()

    @abstractmethod
    def delete(self, *args, **kwargs) -> None:
        raise NotImplementedError()

    @abstractmethod
    def find(self, *args, **kwargs) -> List[T]:
        raise NotImplementedError()

    @abstractmethod
    def find_one(self, *args, **kwargs) -> Optional[T]:
        raise NotImplementedError()

    @abstractmethod
    def get(self, *args, **kwargs) -> T:
        raise NotImplementedError()

    @abstractmethod
    def update(self, *args, **kwargs) -> T:
        raise NotImplementedError()

    @abstractmethod
    def count(self, *args, **kwargs) -> int:
        raise NotImplementedError()



import responses

from posit.connect.client import Client
from posit.connect.content import ContentItem

from .api import load_mock  # type: ignore


class TestContent:
    @responses.activate
    def test_update(self):
        guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        responses.get(
            f"https://connect.example/__api__/v1/content/{guid}",
            json=load_mock(f"v1/content/{guid}.json"),
        )
        con = Client("12345", "https://connect.example")
        content = con.content.get(guid)
        assert content.guid == guid

        new_name = "New Name"
        responses.patch(
            f"https://connect.example/__api__/v1/content/{guid}",
            json={**load_mock(f"v1/content/{guid}.json"), "name": new_name},
        )

        content.update(name=new_name)
        assert content.name == new_name


class TestContents:
    @responses.activate
    def test_get_all_content(self):
        responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )
        con = Client("12345", "https://connect.example")
        all_content = con.content.find()
        assert len(all_content) == 3
        assert [c.name for c in all_content] == [
            "team-admin-dashboard",
            "Performance-Data-1671216053560",
            "My-Streamlit-app",
        ]

    @responses.activate
    def test_content_find_one(self):
        responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )
        con = Client("12345", "https://connect.example")

        one = con.content.find_one()
        assert isinstance(one, ContentItem)
        assert one.name == "team-admin-dashboard"

    @responses.activate
    def test_content_get(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )
        con = Client("12345", "https://connect.example")
        get_one = con.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        assert get_one.name == "Performance-Data-1671216053560"

    @responses.activate
    def test_count(self):
        responses.get(
            "https://connect.example/__api__/v1/content",
            json=load_mock("v1/content.json"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        count = con.content.count()
        assert count == 3



import os

from typing import Optional

from . import urls


def _get_api_key() -> str:
    """Gets the API key from the environment variable 'CONNECT_API_KEY'.

    Raises:
        ValueError: if CONNECT_API_KEY is not set or invalid

    Returns:
        The API key
    """
    value = os.environ.get("CONNECT_API_KEY")
    if not value:
        raise ValueError(
            "Invalid value for 'CONNECT_API_KEY': Must be a non-empty string."
        )
    return value


def _get_url() -> str:
    """Gets the endpoint from the environment variable 'CONNECT_SERVER'.

    The `requests` library uses 'endpoint' instead of 'server'. We will use 'endpoint' from here forward for consistency.

    Raises:
        ValueError: if CONNECT_SERVER is not set or invalid.

    Returns:
        The endpoint.
    """
    value = os.environ.get("CONNECT_SERVER")
    if not value:
        raise ValueError(
            "Invalid value for 'CONNECT_SERVER': Must be a non-empty string."
        )
    return value


class Config:
    """Derived configuration properties"""

    def __init__(
        self, api_key: Optional[str] = None, url: Optional[str] = None
    ) -> None:
        self.api_key = api_key or _get_api_key()
        self.url = urls.server_to_api_url(url or _get_url())
        urls.validate(self.url)



import pytest

from posit.connect.errors import ClientError


class TestClientError:
    def test(self):
        error_code = 0
        error_message = "foo"
        http_status = 404
        http_message = "Foo Bar"
        with pytest.raises(
            ClientError, match=r"foo \(Error Code: 0, HTTP Status: 404 Foo Bar\)"
        ):
            raise ClientError(
                error_code=error_code,
                error_message=error_message,
                http_status=http_status,
                http_message=http_message,
            )



from unittest.mock import MagicMock, Mock, patch

from posit.connect.auth import Auth


class TestAuth:
    @patch("posit.connect.auth.Config")
    def test_auth_headers(self, Config: MagicMock):
        config = Config.return_value
        config.api_key = "foobar"
        auth = Auth(config=config)
        r = Mock()
        r.headers = {}
        auth(r)
        assert r.headers == {"Authorization": f"Key {config.api_key}"}



from http.client import responses
from requests import JSONDecodeError, Response

from .errors import ClientError


def handle_errors(response: Response, *args, **kwargs) -> Response:
    if response.status_code >= 400:
        try:
            data = response.json()
            error_code = data["code"]
            message = data["error"]
            http_status = response.status_code
            http_status_message = responses[http_status]
            raise ClientError(error_code, message, http_status, http_status_message)
        except JSONDecodeError:
            # No JSON error message from Connect, so just raise
            response.raise_for_status()
    return response



import pytest

from posit.connect.urls import append_path, server_to_api_url, validate


def test_append_path():
    assert append_path("http://foo.bar/__api__", "baz") == "http://foo.bar/__api__/baz"


def test_append_path_with_leading_slash():
    assert append_path("http://foo.bar/__api__", "/baz") == "http://foo.bar/__api__/baz"


def test_fix_with_correct_url():
    assert server_to_api_url("http://foo.bar/__api__") == "http://foo.bar/__api__"
    assert server_to_api_url("http://foo.bar/__api__/") == "http://foo.bar/__api__"


def test_fix_without_path():
    assert server_to_api_url("http://foo.bar") == "http://foo.bar/__api__"


def test_fix_with_proxy_path():
    assert server_to_api_url("http://foo.bar/baz") == "http://foo.bar/baz/__api__"


def test_validate_without_scheme():
    with pytest.raises(ValueError):
        validate("foo.bar/__api__")


def test_validate_without_netloc():
    with pytest.raises(ValueError):
        validate("http:///__api__")



# -*- coding: utf-8 -*-
# mypy: ignore-errors
import os

import pandas as pd

from posit.connect.external.databricks import viewer_credentials_provider

from databricks import sql
from databricks.sdk.service.iam import CurrentUserAPI
from databricks.sdk.core import ApiClient, Config

from shiny import App, Inputs, Outputs, Session, render, ui

DATABRICKS_HOST = os.getenv("DATABRICKS_HOST")
DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
SQL_HTTP_PATH = os.getenv("DATABRICKS_PATH")

app_ui = ui.page_fluid(ui.output_text("text"), ui.output_data_frame("result"))


def server(input: Inputs, output: Outputs, session: Session):
    """
    Shiny for Python example application that shows user information and
    the first few rows from a table hosted in Databricks.
    """

    session_token = session.http_conn.headers.get("Posit-Connect-User-Session-Token")
    credentials_provider = viewer_credentials_provider(user_session_token=session_token)

    @render.data_frame
    def result():
        query = "SELECT * FROM samples.nyctaxi.trips LIMIT 10;"

        with sql.connect(
            server_hostname=DATABRICKS_HOST,
            http_path=SQL_HTTP_PATH,
            auth_type="databricks-oauth",
            credentials_provider=credentials_provider,
        ) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query)
                rows = cursor.fetchall()
                df = pd.DataFrame(rows, columns=[col[0] for col in cursor.description])
                return df

    @render.text
    def text():
        cfg = Config(
            host=DATABRICKS_HOST_URL, credentials_provider=credentials_provider
        )
        databricks_user_info = CurrentUserAPI(ApiClient(cfg)).me()
        return f"Hello, {databricks_user_info.display_name}!"


app = App(app_ui, server)



import responses

from posit.connect import Client


class TestOAuthIntegrations:
    @responses.activate
    def test_get_credentials(self):
        responses.post(
            "https://connect.example/__api__/v1/oauth/integrations/credentials",
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                        "subject_token_type": "urn:posit:connect:user-session-token",
                        "subject_token": "cit",
                    }
                )
            ],
            json={
                "access_token": "viewer-token",
                "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
                "token_type": "Bearer",
            },
        )
        responses.post(
            "https://connect.example/__api__/v1/oauth/integrations/credentials",
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                        "subject_token_type": "urn:posit:connect:api-key",
                        "subject_token": "12345",
                    }
                )
            ],
            json={
                "access_token": "sdk-user-token",
                "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
                "token_type": "Bearer",
            },
        )
        con = Client(api_key="12345", url="https://connect.example/")
        assert con.oauth.get_credentials()["access_token"] == "sdk-user-token"
        assert con.oauth.get_credentials("cit")["access_token"] == "viewer-token"



# -*- coding: utf-8 -*-
# mypy: ignore-errors
import os

from posit.connect.external.databricks import viewer_credentials_provider

from databricks import sql
from databricks.sdk.service.iam import CurrentUserAPI
from databricks.sdk.core import ApiClient, Config

from dash import Dash, html, Output, Input, dash_table
import pandas as pd
import flask

DATABRICKS_HOST = os.getenv("DATABRICKS_HOST")
DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
SQL_HTTP_PATH = os.getenv("DATABRICKS_PATH")

df = None
app = Dash(__name__)

app.layout = html.Div(
    children=[
        html.Div(id="greeting", children="Loading..."),
        html.Div(id="table-container"),
        html.Div(id="dummy"),  # dummy element to trigger callback on page load
    ]
)


@app.callback(
    [Output("table-container", "children"), Output("greeting", "children")],
    Input("dummy", "children"),
)
def update_page(_):
    """
    Dash example application that shows user information and
    the first few rows from a table hosted in Databricks.
    """

    session_token = flask.request.headers.get("Posit-Connect-User-Session-Token")
    credentials_provider = viewer_credentials_provider(user_session_token=session_token)

    def get_greeting():
        cfg = Config(
            host=DATABRICKS_HOST_URL, credentials_provider=credentials_provider
        )
        databricks_user_info = CurrentUserAPI(ApiClient(cfg)).me()
        return f"Hello, {databricks_user_info.display_name}!"

    def get_table():
        global df

        if df is None:
            query = "SELECT * FROM samples.nyctaxi.trips LIMIT 10;"

            with sql.connect(
                server_hostname=DATABRICKS_HOST,
                http_path=SQL_HTTP_PATH,
                auth_type="databricks-oauth",
                credentials_provider=credentials_provider,
            ) as connection:
                with connection.cursor() as cursor:
                    cursor.execute(query)
                    rows = cursor.fetchall()
                    df = pd.DataFrame(
                        rows, columns=[col[0] for col in cursor.description]
                    )

        table = dash_table.DataTable(
            id="table",
            columns=[{"name": i, "id": i} for i in df.columns],
            data=df.to_dict("records"),
            style_table={"overflowX": "scroll"},
        )
        return table

    return get_table(), get_greeting()


if __name__ == "__main__":
    app.run(debug=True)



from __future__ import annotations

from requests import Response, Session
from typing import Optional

from . import hooks, me, urls

from .auth import Auth
from .config import Config
from .oauth import OAuthIntegration
from .content import Content
from .users import User, Users


class Client:
    def __init__(
        self,
        api_key: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        """
        Initialize the Client instance.

        Args:
            api_key (str, optional): API key for authentication. Defaults to None.
            url (str, optional): API url URL. Defaults to None.
        """
        # Create a Config object.
        self.config = Config(api_key=api_key, url=url)
        # Create a Session object for making HTTP requests.
        session = Session()
        # Authenticate the session using the provided Config.
        session.auth = Auth(config=self.config)
        # Add error handling hooks to the session.
        session.hooks["response"].append(hooks.handle_errors)

        # Store the Session object.
        self.session = session

        # Internal attributes to hold settings we fetch lazily
        self._server_settings = None

    @property
    def connect_version(self):
        if self._server_settings is None:
            self._server_settings = self.get("server_settings").json()
        return self._server_settings["version"]

    @property
    def me(self) -> User:
        return me.get(self.config, self.session)

    @property
    def oauth(self) -> OAuthIntegration:
        return OAuthIntegration(config=self.config, session=self.session)

    @property
    def users(self) -> Users:
        return Users(config=self.config, session=self.session)

    @property
    def content(self) -> Content:
        return Content(config=self.config, session=self.session)

    def __del__(self):
        """
        Close the session when the Client instance is deleted.
        """
        if hasattr(self, "session") and self.session is not None:
            self.session.close()

    def __enter__(self):
        """
        Enter method for using the client as a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """
        Closes the session if it exists.

        Args:
            exc_type: The type of the exception raised (if any).
            exc_value: The exception instance raised (if any).
            exc_tb: The traceback for the exception raised (if any).
        """
        if hasattr(self, "session") and self.session is not None:
            self.session.close()

    def request(self, method: str, path: str, **kwargs) -> Response:
        """
        Sends an HTTP request to the specified path using the given method.

        Args:
            method (str): The HTTP method to use for the request.
            path (str): The path to send the request to.
            **kwargs: Additional keyword arguments to pass to the underlying session's request method.

        Returns:
            Response: The response object containing the server's response to the request.
        """
        url = urls.append_path(self.config.url, path)
        return self.session.request(method, url, **kwargs)

    def get(self, path: str, **kwargs) -> Response:
        """
        Send a GET request to the specified path.

        Args:
            path (str): The path to send the request to.
            **kwargs: Additional keyword arguments to be passed to the underlying session's `get` method.

        Returns:
            Response: The response object.

        """
        url = urls.append_path(self.config.url, path)
        return self.session.get(url, **kwargs)

    def post(self, path: str, **kwargs) -> Response:
        """
        Send a POST request to the specified path.

        Args:
            path (str): The path to send the request to.
            **kwargs: Additional keyword arguments to be passed to the underlying session's `post` method.

        Returns:
            Response: The response object.

        """
        url = urls.append_path(self.config.url, path)
        return self.session.post(url, **kwargs)

    def put(self, path: str, **kwargs) -> Response:
        """
        Send a PUT request to the specified path.

        Args:
            path (str): The path to send the request to.
            **kwargs: Additional keyword arguments to be passed to the underlying session's `put` method.

        Returns:
            Response: The response object.

        """
        url = urls.append_path(self.config.url, path)
        return self.session.put(url, **kwargs)

    def patch(self, path: str, **kwargs) -> Response:
        """
        Send a PATCH request to the specified path.

        Args:
            path (str): The path to send the request to.
            **kwargs: Additional keyword arguments to be passed to the underlying session's `patch` method.

        Returns:
            Response: The response object.

        """
        url = urls.append_path(self.config.url, path)
        return self.session.patch(url, **kwargs)

    def delete(self, path: str, **kwargs) -> Response:
        """
        Send a DELETE request to the specified path.

        Args:
            path (str): The path to send the request to.
            **kwargs: Additional keyword arguments to be passed to the underlying session's `delete` method.

        Returns:
            Response: The response object.

        """
        url = urls.append_path(self.config.url, path)
        return self.session.delete(url, **kwargs)



from __future__ import annotations

from typing import List, Optional, overload


from requests import Session

from . import urls

from .config import Config
from .resources import Resources, Resource


class ContentItem(Resource):
    @property
    def guid(self) -> str:
        return self.get("guid")  # type: ignore

    @property
    def name(self) -> str:
        return self.get("name")  # type: ignore

    @property
    def title(self) -> Optional[str]:
        return self.get("title")  # type: ignore

    @property
    def description(self) -> str:
        return self.get("description")  # type: ignore

    @property
    def access_type(self) -> str:
        return self.get("access_type")  # type: ignore

    @property
    def connection_timeout(self) -> Optional[int]:
        return self.get("connection_timeout")  # type: ignore

    @property
    def read_timeout(self) -> Optional[int]:
        return self.get("read_timeout")  # type: ignore

    @property
    def init_timeout(self) -> Optional[int]:
        return self.get("init_timeout")  # type: ignore

    @property
    def idle_timeout(self) -> Optional[int]:
        return self.get("idle_timeout")  # type: ignore

    @property
    def max_processes(self) -> Optional[int]:
        return self.get("max_processes")  # type: ignore

    @property
    def min_processes(self) -> Optional[int]:
        return self.get("min_processes")  # type: ignore

    @property
    def max_conns_per_process(self) -> Optional[int]:
        return self.get("max_conns_per_process")  # type: ignore

    @property
    def load_factor(self) -> Optional[float]:
        return self.get("load_factor")  # type: ignore

    @property
    def cpu_request(self) -> Optional[float]:
        return self.get("cpu_request")  # type: ignore

    @property
    def cpu_limit(self) -> Optional[float]:
        return self.get("cpu_limit")  # type: ignore

    @property
    def memory_request(self) -> Optional[int]:
        return self.get("memory_request")  # type: ignore

    @property
    def memory_limit(self) -> Optional[int]:
        return self.get("memory_limit")  # type: ignore

    @property
    def amd_gpu_limit(self) -> Optional[int]:
        return self.get("amd_gpu_limit")  # type: ignore

    @property
    def nvidia_gpu_limit(self) -> Optional[int]:
        return self.get("nvidia_gpu_limit")  # type: ignore

    @property
    def created_time(self) -> str:
        return self.get("created_time")  # type: ignore

    @property
    def last_deployed_time(self) -> str:
        return self.get("last_deployed_time")  # type: ignore

    @property
    def bundle_id(self) -> Optional[str]:
        return self.get("bundle_id")  # type: ignore

    @property
    def app_mode(self) -> str:
        return self.get("app_mode")  # type: ignore

    @property
    def content_category(self) -> Optional[str]:
        return self.get("content_category")  # type: ignore

    @property
    def parameterized(self) -> bool:
        return self.get("parameterized")  # type: ignore

    @property
    def cluster_name(self) -> Optional[str]:
        return self.get("cluster_name")  # type: ignore

    @property
    def image_name(self) -> Optional[str]:
        return self.get("image_name")  # type: ignore

    @property
    def default_image_name(self) -> Optional[str]:
        return self.get("default_image_name")  # type: ignore

    @property
    def default_r_environment_management(self) -> Optional[bool]:
        return self.get("default_r_environment_management")  # type: ignore

    @property
    def default_py_environment_management(self) -> Optional[bool]:
        return self.get("default_py_environment_management")  # type: ignore

    @property
    def service_account_name(self) -> Optional[str]:
        return self.get("service_account_name")  # type: ignore

    @property
    def r_version(self) -> Optional[str]:
        return self.get("r_version")  # type: ignore

    @property
    def r_environment_management(self) -> Optional[bool]:
        return self.get("r_environment_management")  # type: ignore

    @property
    def py_version(self) -> Optional[str]:
        return self.get("py_version")  # type: ignore

    @property
    def py_environment_management(self) -> Optional[bool]:
        return self.get("py_environment_management")  # type: ignore

    @property
    def quarto_version(self) -> Optional[str]:
        return self.get("quarto_version")  # type: ignore

    @property
    def run_as(self) -> Optional[str]:
        return self.get("run_as")  # type: ignore

    @property
    def run_as_current_user(self) -> bool:
        return self.get("run_as_current_user")  # type: ignore

    @property
    def owner_guid(self) -> str:
        return self.get("owner_guid")  # type: ignore

    @property
    def content_url(self) -> str:
        return self.get("content_url")  # type: ignore

    @property
    def dashboard_url(self) -> str:
        return self.get("dashboard_url")  # type: ignore

    @property
    def app_role(self) -> str:
        return self.get("app_role")  # type: ignore

    @property
    def id(self) -> str:
        return self.get("id")  # type: ignore

    @overload
    def update(
        self,
        name: str = ...,
        title: Optional[str] = ...,
        description: str = ...,
        access_type: str = ...,
        owner_guid: Optional[str] = ...,
        connection_timeout: Optional[int] = ...,
        read_timeout: Optional[int] = ...,
        init_timeout: Optional[int] = ...,
        idle_timeout: Optional[int] = ...,
        max_processes: Optional[int] = ...,
        min_processes: Optional[int] = ...,
        max_conns_per_process: Optional[int] = ...,
        load_factor: Optional[float] = ...,
        cpu_request: Optional[float] = ...,
        cpu_limit: Optional[float] = ...,
        memory_request: Optional[int] = ...,
        memory_limit: Optional[int] = ...,
        amd_gpu_limit: Optional[int] = ...,
        nvidia_gpu_limit: Optional[int] = ...,
        run_as: Optional[str] = ...,
        run_as_current_user: Optional[bool] = ...,
        default_image_name: Optional[str] = ...,
        default_r_environment_management: Optional[bool] = ...,
        default_py_environment_management: Optional[bool] = ...,
        service_account_name: Optional[str] = ...,
    ) -> None:
        """
        Update the content item.

        Args:
            name (str): The name of the content.
            title (Optional[str]): The title of the content.
            description (str): The description of the content.
            access_type (str): The access type of the content.
            owner_guid (Optional[str]): The owner GUID of the content.
            connection_timeout (Optional[int]): The connection timeout in seconds.
            read_timeout (Optional[int]): The read timeout in seconds.
            init_timeout (Optional[int]): The initialization timeout in seconds.
            idle_timeout (Optional[int]): The idle timeout in seconds.
            max_processes (Optional[int]): The maximum number of processes.
            min_processes (Optional[int]): The minimum number of processes.
            max_conns_per_process (Optional[int]): The maximum number of connections per process.
            load_factor (Optional[float]): The load factor.
            cpu_request (Optional[float]): The CPU request.
            cpu_limit (Optional[float]): The CPU limit.
            memory_request (Optional[int]): The memory request in bytes.
            memory_limit (Optional[int]): The memory limit in bytes.
            amd_gpu_limit (Optional[int]): The AMD GPU limit.
            nvidia_gpu_limit (Optional[int]): The NVIDIA GPU limit.
            run_as (Optional[str]): The user to run as.
            run_as_current_user (Optional[bool]): Whether to run as the current user.
            default_image_name (Optional[str]): The default image name.
            default_r_environment_management (Optional[bool]): Whether to use default R environment management.
            default_py_environment_management (Optional[bool]): Whether to use default Python environment management.
            service_account_name (Optional[str]): The service account name.

        Returns:
            None
        """
        ...

    @overload
    def update(self, *args, **kwargs) -> None:
        """
        Update the content item.

        Args:
            *args
            **kwargs

        Returns:
            None
        """
        ...

    def update(self, *args, **kwargs) -> None:
        """
        Update the content item.

        Args:
            *args
            **kwargs

        Returns:
            None
        """
        body = dict(*args, **kwargs)
        url = urls.append_path(self.config.url, f"v1/content/{self.guid}")
        response = self.session.patch(url, json=body)
        super().update(**response.json())


class Content(Resources[ContentItem]):
    def __init__(self, config: Config, session: Session) -> None:
        self.url = urls.append_path(config.url, "v1/content")
        self.config = config
        self.session = session

    def find(self) -> List[ContentItem]:
        results = self.session.get(self.url).json()
        items = (
            ContentItem(
                config=self.config,
                session=self.session,
                **result,
            )
            for result in results
        )
        return [item for item in items]

    def find_one(self) -> ContentItem | None:
        results = self.session.get(self.url).json()
        items = (
            ContentItem(
                config=self.config,
                session=self.session,
                **result,
            )
            for result in results
        )
        return next(items, None)

    def get(self, id: str) -> ContentItem:
        url = urls.append_path(self.url, id)
        response = self.session.get(url)
        return ContentItem(self.config, self.session, **response.json())

    def create(self) -> ContentItem:
        raise NotImplementedError()

    def update(self) -> ContentItem:
        raise NotImplementedError()

    def delete(self) -> None:
        raise NotImplementedError()

    def count(self) -> int:
        results = self.session.get(self.url).json()
        return len(results)



from dataclasses import dataclass
from typing import Generator, List

import requests


# The maximum page size supported by the API.
_MAX_PAGE_SIZE = 500


@dataclass
class Page:
    """
    Represents a page of results returned by the paginator.

    Attributes:
        current_page (int): The current page number.
        total (int): The total number of results.
        results (List[dict]): The list of results on the current page.
    """

    current_page: int
    total: int
    results: List[dict]


class Paginator:
    """
    A class for paginating through API results.

    Args:
        session (requests.Session): The session object to use for making API requests.
        url (str): The URL of the paginated API endpoint.

    Attributes:
        session (requests.Session): The session object to use for making API requests.
        url (str): The URL of the paginated API endpoint.
    """

    def __init__(self, session: requests.Session, url: str, params: dict = {}) -> None:
        self.session = session
        self.url = url
        self.params = params

    def fetch_results(self) -> List[dict]:
        """
        Fetches and returns all the results from the paginated API endpoint.

        Returns:
            A list of dictionaries representing the fetched results.
        """
        results = []
        for page in self.fetch_pages():
            results.extend(page.results)
        return results

    def fetch_pages(self) -> Generator[Page, None, None]:
        """
        Fetches pages of results from the API.

        Yields:
            Page: A page of results from the API.
        """
        count = 0
        page_number = 1
        while True:
            page = self.fetch_page(page_number)
            page_number += 1
            if len(page.results) > 0:
                yield page
            else:
                # stop if the result set is empty
                return

            count += len(page.results)
            # Check if the local count has reached the total threshold.
            # It is possible for count to exceed total if the total changes
            # during execution of this loop.
            # It is also possible for the total to change between iterations.
            if count >= page.total:
                break

    def fetch_page(self, page_number: int) -> Page:
        """
        Fetches a specific page of data from the API.

        Args:
            page_number (int): The page number to fetch.

        Returns:
            Page: The fetched page object.

        """
        params = {
            **self.params,
            "page_number": page_number,
            "page_size": _MAX_PAGE_SIZE,
        }
        response = self.session.get(self.url, params=params)
        return Page(**response.json())



# -*- coding: utf-8 -*-
# mypy: ignore-errors
import os

from posit.connect.external.databricks import viewer_credentials_provider

from databricks import sql
from databricks.sdk.service.iam import CurrentUserAPI
from databricks.sdk.core import ApiClient, Config

import pandas as pd
import streamlit as st
from streamlit.web.server.websocket_headers import _get_websocket_headers

DATABRICKS_HOST = os.getenv("DATABRICKS_HOST")
DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
SQL_HTTP_PATH = os.getenv("DATABRICKS_PATH")

session_token = _get_websocket_headers().get("Posit-Connect-User-Session-Token")

credentials_provider = viewer_credentials_provider(user_session_token=session_token)

cfg = Config(host=DATABRICKS_HOST_URL, credentials_provider=credentials_provider)
databricks_user = CurrentUserAPI(ApiClient(cfg)).me()
st.write(f"Hello, {databricks_user.display_name}!")

with sql.connect(
    server_hostname=DATABRICKS_HOST,
    http_path=SQL_HTTP_PATH,
    auth_type="databricks-oauth",
    credentials_provider=credentials_provider,
) as connection:
    with connection.cursor() as cursor:
        cursor.execute("SELECT * FROM samples.nyctaxi.trips LIMIT 10;")
        result = cursor.fetchall()
        st.table(pd.DataFrame(result))



# -*- coding: utf-8 -*-
# mypy: ignore-errors
import os

from posit.connect.external.databricks import viewer_credentials_provider

from databricks import sql

from typing import Annotated

from fastapi import FastAPI, Header
from fastapi.responses import JSONResponse

DATABRICKS_HOST = os.getenv("DATABRICKS_HOST")
DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
SQL_HTTP_PATH = os.getenv("DATABRICKS_PATH")

rows = None
app = FastAPI()


@app.get("/fares")
async def get_fares(
    posit_connect_user_session_token: Annotated[str | None, Header()] = None,
) -> JSONResponse:
    """
    FastAPI example API that returns the first few rows from
    a table hosted in Databricks.
    """
    global rows

    credentials_provider = viewer_credentials_provider(
        user_session_token=posit_connect_user_session_token
    )

    if rows is None:
        query = "SELECT * FROM samples.nyctaxi.trips LIMIT 10;"

        with sql.connect(
            server_hostname=DATABRICKS_HOST,
            http_path=SQL_HTTP_PATH,
            auth_type="databricks-oauth",
            credentials_provider=credentials_provider,
        ) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query)
                rows = cursor.fetchall()

    return [row.asDict() for row in rows]



import pytest
import responses

from unittest.mock import MagicMock, patch

from posit.connect import Client

from .api import load_mock  # type: ignore


@pytest.fixture
def MockAuth():
    with patch("posit.connect.client.Auth") as mock:
        yield mock


@pytest.fixture
def MockConfig():
    with patch("posit.connect.client.Config") as mock:
        yield mock


@pytest.fixture
def MockSession():
    with patch("posit.connect.client.Session") as mock:
        yield mock


class TestClient:
    def test_init(
        self,
        MockAuth: MagicMock,
        MockConfig: MagicMock,
        MockSession: MagicMock,
    ):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        Client(api_key=api_key, url=url)
        MockAuth.assert_called_once_with(config=MockConfig.return_value)
        MockConfig.assert_called_once_with(api_key=api_key, url=url)
        MockSession.assert_called_once()

    def test__del__(self, MockAuth, MockConfig, MockSession):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        client = Client(api_key=api_key, url=url)
        del client
        MockSession.return_value.close.assert_called_once()

    def test__enter__(self):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        with Client(api_key=api_key, url=url) as client:
            assert isinstance(client, Client)

    def test__exit__(self, MockSession):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        with Client(api_key=api_key, url=url) as client:
            assert isinstance(client, Client)
        MockSession.return_value.close.assert_called_once()

    @responses.activate
    def test_connect_version(self):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        client = Client(api_key=api_key, url=url)

        # The actual server_settings response has a lot more attributes, but we
        # don't need to include them all here because we don't use them
        responses.get(
            "http://foo.bar/__api__/server_settings",
            json={"version": "2024.01.0"},
        )
        assert client.connect_version == "2024.01.0"

    @responses.activate
    def test_me_request(self):
        responses.get(
            "https://connect.example/__api__/v1/user",
            json=load_mock("v1/user.json"),
        )

        con = Client(api_key="12345", url="https://connect.example/")
        assert con.me.username == "carlos12"

    def test_request(self, MockSession):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        client = Client(api_key=api_key, url=url)
        client.request("GET", "/foo")
        MockSession.return_value.request.assert_called_once_with(
            "GET", "http://foo.bar/__api__/foo"
        )

    def test_get(self, MockSession):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        client = Client(api_key=api_key, url=url)
        client.get("/foo")
        client.session.get.assert_called_once_with("http://foo.bar/__api__/foo")

    def test_post(self, MockSession):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        client = Client(api_key=api_key, url=url)
        client.post("/foo")
        client.session.post.assert_called_once_with("http://foo.bar/__api__/foo")

    def test_put(self, MockSession):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        client = Client(api_key=api_key, url=url)
        client.put("/foo")
        client.session.put.assert_called_once_with("http://foo.bar/__api__/foo")

    def test_patch(self, MockSession):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        client = Client(api_key=api_key, url=url)
        client.patch("/foo")
        client.session.patch.assert_called_once_with("http://foo.bar/__api__/foo")

    def test_delete(self, MockSession):
        api_key = "foobar"
        url = "http://foo.bar/__api__"
        client = Client(api_key=api_key, url=url)
        client.delete("/foo")



class ClientError(Exception):
    def __init__(
        self, error_code: int, error_message: str, http_status: int, http_message: str
    ):
        self.error_code = error_code
        self.error_message = error_message
        self.http_status = http_status
        self.http_message = http_message
        super().__init__(
            f"{error_message} (Error Code: {error_code}, HTTP Status: {http_status} {http_message})"
        )



from __future__ import annotations

from requests import Session
from typing import Optional, TypedDict

from . import urls
from .config import Config


class Credentials(TypedDict, total=False):
    access_token: str
    issued_token_type: str
    token_type: str


class OAuthIntegration:
    def __init__(self, config: Config, session: Session) -> None:
        self.url = urls.append_path(config.url, "v1/oauth/integrations/credentials")
        self.config = config
        self.session = session

    def get_credentials(self, user_session_token: Optional[str] = None) -> Credentials:
        # craft a basic credential exchange request where the self.config.api_key owner
        # is requesting their own credentials
        data = dict()
        data["grant_type"] = "urn:ietf:params:oauth:grant-type:token-exchange"
        data["subject_token_type"] = "urn:posit:connect:api-key"
        data["subject_token"] = self.config.api_key

        # if this content is running on Connect, then it is allowed to request
        # the content viewer's credentials
        if user_session_token:
            data["subject_token_type"] = "urn:posit:connect:user-session-token"
            data["subject_token"] = user_session_token

        response = self.session.post(self.url, data=data)
        return Credentials(**response.json())



import requests

from . import urls

from .config import Config
from .users import User


def get(config: Config, session: requests.Session) -> User:
    """
    Gets the current user.

    Args:
        config (Config): The configuration object containing the URL.
        session (requests.Session): The session object used for making HTTP requests.

    Returns:
        User: The current user.
    """
    url = urls.append_path(config.url, "v1/user")
    response = session.get(url)
    return User(config, session, **response.json())



from unittest.mock import Mock

import pytest
import requests
import responses

from posit.connect.client import Client
from posit.connect.users import User

from .api import load_mock  # type: ignore

session = Mock()
url = Mock()


class TestUserAttributes:
    def test_guid(self):
        user = User(session, url)
        assert hasattr(user, "guid")
        assert user.guid is None
        user = User(session, url, guid="test_guid")
        assert user.guid == "test_guid"

    def test_email(self):
        user = User(session, url)
        assert hasattr(user, "email")
        assert user.email is None
        user = User(session, url, email="test@example.com")
        assert user.email == "test@example.com"

    def test_username(self):
        user = User(session, url)
        assert hasattr(user, "username")
        assert user.username is None
        user = User(session, url, username="test_user")
        assert user.username == "test_user"

    def test_first_name(self):
        user = User(session, url)
        assert hasattr(user, "first_name")
        assert user.first_name is None
        user = User(session, url, first_name="John")
        assert user.first_name == "John"

    def test_last_name(self):
        user = User(session, url)
        assert hasattr(user, "last_name")
        assert user.last_name is None
        user = User(session, url, last_name="Doe")
        assert user.last_name == "Doe"

    def test_user_role(self):
        user = User(session, url)
        assert hasattr(user, "user_role")
        assert user.user_role is None
        user = User(session, url, user_role="admin")
        assert user.user_role == "admin"

    def test_created_time(self):
        user = User(session, url)
        assert hasattr(user, "created_time")
        assert user.created_time is None
        user = User(session, url, created_time="2022-01-01T00:00:00")
        assert user.created_time == "2022-01-01T00:00:00"

    def test_updated_time(self):
        user = User(session, url)
        assert hasattr(user, "updated_time")
        assert user.updated_time is None
        user = User(session, url, updated_time="2022-01-01T00:00:00")
        assert user.updated_time == "2022-01-01T00:00:00"

    def test_active_time(self):
        user = User(session, url)
        assert hasattr(user, "active_time")
        assert user.active_time is None
        user = User(session, url, active_time="2022-01-01T00:00:00")
        assert user.active_time == "2022-01-01T00:00:00"

    def test_confirmed(self):
        user = User(session, url)
        assert hasattr(user, "confirmed")
        assert user.confirmed is None
        user = User(session, url, confirmed=True)
        assert user.confirmed is True

    def test_locked(self):
        user = User(session, url)
        assert hasattr(user, "locked")
        assert user.locked is None
        user = User(session, url, locked=False)
        assert user.locked is False


class TestUserLock:
    @responses.activate
    def test_lock(self):
        responses.get(
            "https://connect.example/__api__/v1/users/a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6",
            json=load_mock("v1/users/a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6.json"),
        )
        c = Client(api_key="12345", url="https://connect.example/")
        user = c.users.get("a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6")
        assert user.guid == "a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6"

        responses.get(
            "https://connect.example/__api__/v1/user",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        responses.post(
            "https://connect.example/__api__/v1/users/a1b2c3d4-e5f6-g7h8-i9j0-k1l2m3n4o5p6/lock",
            match=[responses.matchers.json_params_matcher({"locked": True})],
        )
        user.lock()
        assert user.locked

    @responses.activate
    def test_lock_self_true(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        c = Client(api_key="12345", url="https://connect.example/")
        user = c.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert user.guid == "20a79ce3-6e87-4522-9faf-be24228800a4"

        responses.get(
            "https://connect.example/__api__/v1/user",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        responses.post(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4/lock",
            match=[responses.matchers.json_params_matcher({"locked": True})],
        )
        user.lock(force=True)
        assert user.locked

    @responses.activate
    def test_lock_self_false(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        c = Client(api_key="12345", url="https://connect.example/")
        user = c.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert user.guid == "20a79ce3-6e87-4522-9faf-be24228800a4"

        responses.get(
            "https://connect.example/__api__/v1/user",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        responses.post(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4/lock",
            match=[responses.matchers.json_params_matcher({"locked": True})],
        )
        with pytest.raises(RuntimeError):
            user.lock(force=False)
        assert not user.locked


class TestUserUnlock:
    @responses.activate
    def test_unlock(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        c = Client(api_key="12345", url="https://connect.example/")
        user = c.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert user.guid == "20a79ce3-6e87-4522-9faf-be24228800a4"

        responses.post(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4/lock",
            match=[responses.matchers.json_params_matcher({"locked": False})],
        )
        user.unlock()
        assert not user.locked


class TestUsers:
    @responses.activate
    def test_users_get(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )

        con = Client(api_key="12345", url="https://connect.example/")
        carlos = con.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert carlos.username == "carlos12"
        assert carlos.first_name == "Carlos"
        assert carlos.created_time == "2019-09-09T15:24:32Z"

    @responses.activate
    def test_users_get_extra_fields(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json={
                "guid": "20a79ce3-6e87-4522-9faf-be24228800a4",
                "username": "carlos12",
                "some_new_field": "some_new_value",
            },
        )

        con = Client(api_key="12345", url="https://connect.example/")
        carlos = con.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        assert carlos.username == "carlos12"
        assert carlos["some_new_field"] == "some_new_value"

    @responses.activate
    def test_user_update(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        patch_request = responses.put(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            match=[responses.matchers.json_params_matcher({"first_name": "Carlitos"})],
            json={"first_name": "Carlitos"},
        )

        con = Client(api_key="12345", url="https://connect.example/")
        carlos = con.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")

        assert patch_request.call_count == 0
        assert carlos.first_name == "Carlos"

        carlos.update(first_name="Carlitos")

        assert patch_request.call_count == 1
        assert carlos.first_name == "Carlitos"

    @responses.activate
    def test_user_update_server_error(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )
        responses.put(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            status=500,
        )

        con = Client(api_key="12345", url="https://connect.example/")
        carlos = con.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")
        with pytest.raises(requests.HTTPError, match="500 Server Error"):
            carlos.update(first_name="Carlitos")

    @responses.activate
    def test_user_cant_setattr(self):
        responses.get(
            "https://connect.example/__api__/v1/users/20a79ce3-6e87-4522-9faf-be24228800a4",
            json=load_mock("v1/users/20a79ce3-6e87-4522-9faf-be24228800a4.json"),
        )

        con = Client(api_key="12345", url="https://connect.example/")
        carlos = con.users.get("20a79ce3-6e87-4522-9faf-be24228800a4")

        with pytest.raises(
            AttributeError,
            match=r"cannot set attributes: use update\(\) instead",
        ):
            carlos.first_name = "Carlitos"

    @responses.activate
    def test_count(self):
        responses.get(
            "https://connect.example/__api__/v1/users",
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
            match=[responses.matchers.query_param_matcher({"page_size": 1})],
        )
        con = Client(api_key="12345", url="https://connect.example/")
        count = con.users.count()
        assert count == 3


class TestUsersFindOne:
    @responses.activate
    def test_default(self):
        # validate first result returned
        responses.get(
            "https://connect.example/__api__/v1/users",
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        user = con.users.find_one()
        assert user.username == "al"
        assert len(responses.calls) == 1

    @responses.activate
    def test_params(self):
        # validate input params are propagated to the query params
        params = {"key1": "value1", "key2": "value2", "key3": "value3"}
        responses.get(
            "https://connect.example/__api__/v1/users",
            match=[
                responses.matchers.query_param_matcher(
                    {"page_size": 500, "page_number": 1, **params}
                )
            ],
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        con.users.find_one(**params)
        responses.assert_call_count(
            "https://connect.example/__api__/v1/users?key1=value1&key2=value2&key3=value3&page_number=1&page_size=500",
            1,
        )

    @responses.activate
    def test_empty_results(self):
        responses.get(
            "https://connect.example/__api__/v1/users",
            json={"total": 0, "current_page": 1, "results": []},
        )

        con = Client(api_key="12345", url="https://connect.example/")
        user = con.users.find_one()
        assert user is None


class TestUsersFind:
    @responses.activate
    def test_default(self):
        # validate response body is parsed and returned
        responses.get(
            "https://connect.example/__api__/v1/users",
            match=[
                responses.matchers.query_param_matcher(
                    {"page_size": 500, "page_number": 1}
                )
            ],
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
        )
        responses.get(
            "https://connect.example/__api__/v1/users",
            match=[
                responses.matchers.query_param_matcher(
                    {"page_size": 500, "page_number": 2}
                )
            ],
            json=load_mock("v1/users?page_number=2&page_size=500.jsonc"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        users = con.users.find()
        assert len(users) == 3
        assert users[0] == {
            "email": "alice@connect.example",
            "username": "al",
            "first_name": "Alice",
            "last_name": "User",
            "user_role": "publisher",
            "created_time": "2017-08-08T15:24:32Z",
            "updated_time": "2023-03-02T20:25:06Z",
            "active_time": "2018-05-09T16:58:45Z",
            "confirmed": True,
            "locked": False,
            "guid": "a01792e3-2e67-402e-99af-be04a48da074",
        }

    @responses.activate
    def test_params(self):
        # validate input params are propagated to the query params
        params = {"key1": "value1", "key2": "value2", "key3": "value3"}
        responses.get(
            "https://connect.example/__api__/v1/users",
            match=[
                responses.matchers.query_param_matcher(
                    {"page_size": 500, "page_number": 1, **params}
                )
            ],
            json=load_mock("v1/users?page_number=1&page_size=500.jsonc"),
        )
        responses.get(
            "https://connect.example/__api__/v1/users",
            match=[
                responses.matchers.query_param_matcher(
                    {"page_size": 500, "page_number": 2, **params}
                )
            ],
            json=load_mock("v1/users?page_number=2&page_size=500.jsonc"),
        )
        con = Client(api_key="12345", url="https://connect.example/")
        con.users.find(**params)
        responses.assert_call_count(
            "https://connect.example/__api__/v1/users?key1=value1&key2=value2&key3=value3&page_number=1&page_size=500",
            1,
        )

    @responses.activate
    def test_params_not_dict_like(self):
        # validate input params are propagated to the query params
        con = Client(api_key="12345", url="https://connect.example/")
        not_dict_like = "string"
        with pytest.raises(ValueError):
            con.users.find(not_dict_like)



import abc
import os
from typing import Callable, Dict, Optional

from ..client import Client
from ..oauth import OAuthIntegration

HeaderFactory = Callable[[], Dict[str, str]]


# https://github.com/databricks/databricks-sdk-py/blob/v0.20.0/databricks/sdk/credentials_provider.py
# https://github.com/databricks/databricks-sql-python/blob/v3.1.0/src/databricks/sql/auth/authenticators.py
# In order to keep compatibility with the Databricks SDK
class CredentialsProvider(abc.ABC):
    """CredentialsProvider is the protocol (call-side interface)
    for authenticating requests to Databricks REST APIs"""

    @abc.abstractmethod
    def auth_type(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self, *args, **kwargs) -> HeaderFactory:
        raise NotImplementedError


class PositOAuthIntegrationCredentialsProvider(CredentialsProvider):
    def __init__(self, posit_oauth: OAuthIntegration, user_session_token: str):
        self.posit_oauth = posit_oauth
        self.user_session_token = user_session_token

    def auth_type(self) -> str:
        return "posit-oauth-integration"

    def __call__(self, *args, **kwargs) -> HeaderFactory:
        def inner() -> Dict[str, str]:
            access_token = self.posit_oauth.get_credentials(self.user_session_token)[
                "access_token"
            ]
            return {"Authorization": f"Bearer {access_token}"}

        return inner


# Use this environment variable to determine if the
# client SDK was initialized from a piece of content running on a Connect server.
def is_local() -> bool:
    return not os.getenv("RSTUDIO_PRODUCT") == "CONNECT"


def viewer_credentials_provider(
    client: Optional[Client] = None, user_session_token: Optional[str] = None
) -> Optional[CredentialsProvider]:
    # If the content is not running on Connect then viewer auth should
    # fall back to the locally configured credentials hierarchy
    if is_local():
        return None

    if client is None:
        client = Client()

    # If the user-session-token wasn't provided and we're running on Connect then we raise an exception.
    # user_session_token is required to impersonate the viewer.
    if user_session_token is None:
        raise ValueError(
            "The user-session-token is required for viewer authentication."
        )

    return PositOAuthIntegrationCredentialsProvider(client.oauth, user_session_token)


def service_account_credentials_provider(client: Optional[Client] = None):
    raise NotImplementedError



from __future__ import annotations
from typing import List, overload


import requests


from . import me, urls

from .config import Config
from .paginator import Paginator
from .resources import Resource, Resources


class User(Resource):
    @property
    def guid(self) -> str:
        return self.get("guid")  # type: ignore

    @property
    def email(self) -> str:
        return self.get("email")  # type: ignore

    @property
    def username(self) -> str:
        return self.get("username")  # type: ignore

    @property
    def first_name(self) -> str:
        return self.get("first_name")  # type: ignore

    @property
    def last_name(self) -> str:
        return self.get("last_name")  # type: ignore

    @property
    def user_role(self) -> str:
        return self.get("user_role")  # type: ignore

    @property
    def created_time(self) -> str:
        return self.get("created_time")  # type: ignore

    @property
    def updated_time(self) -> str:
        return self.get("updated_time")  # type: ignore

    @property
    def active_time(self) -> str:
        return self.get("active_time")  # type: ignore

    @property
    def confirmed(self) -> bool:
        return self.get("confirmed")  # type: ignore

    @property
    def locked(self) -> bool:
        return self.get("locked")  # type: ignore

    def lock(self, *, force: bool = False):
        """
        Locks the user account.

        .. warning:: You cannot unlock your own account. Once an account is locked, only an admin can unlock it.

        Args:
            force (bool, optional): If set to True, overrides lock protection allowing a user to lock their own account. Defaults to False.

        Returns:
            None
        """
        _me = me.get(self.config, self.session)
        if _me.guid == self.guid and not force:
            raise RuntimeError(
                "You cannot lock your own account. Set force=True to override this behavior."
            )
        url = urls.append_path(self.config.url, f"v1/users/{self.guid}/lock")
        body = {"locked": True}
        self.session.post(url, json=body)
        super().update(locked=True)

    def unlock(self):
        """
        Unlocks the user account.

        Returns:
            None
        """
        url = urls.append_path(self.config.url, f"v1/users/{self.guid}/lock")
        body = {"locked": False}
        self.session.post(url, json=body)
        super().update(locked=False)

    @overload
    def update(
        self,
        email: str = ...,
        username: str = ...,
        first_name: str = ...,
        last_name: str = ...,
        user_role: str = ...,
    ) -> None:
        """
        Update the user.

        Args:
            email (str): The email address for the user.
            username (str): The username for the user.
            first_name (str): The first name for the user.
            last_name (str): The last name for the user.
            user_role (str): The role for the user.

        Returns:
            None
        """
        ...

    @overload
    def update(self, *args, **kwargs) -> None:
        """
        Update the user.

        Args:
            *args
            **kwargs

        Returns:
            None
        """
        ...

    def update(self, *args, **kwargs) -> None:
        """
        Update the user.

        Args:
            *args
            **kwargs

        Returns:
            None
        """
        body = dict(*args, **kwargs)
        url = urls.append_path(self.config.url, f"v1/users/{self.guid}")
        response = self.session.put(url, json=body)
        super().update(**response.json())


class Users(Resources[User]):
    def __init__(self, config: Config, session: requests.Session) -> None:
        self.url = urls.append_path(config.url, "v1/users")
        self.config = config
        self.session = session

    @overload
    def find(
        self, prefix: str = ..., user_role: str = ..., account_status: str = ...
    ) -> List[User]: ...

    @overload
    def find(self, *args, **kwargs) -> List[User]: ...

    def find(self, *args, **kwargs):
        params = dict(*args, **kwargs)
        paginator = Paginator(self.session, self.url, params=params)
        results = paginator.fetch_results()
        return [
            User(
                config=self.config,
                session=self.session,
                **user,
            )
            for user in results
        ]

    @overload
    def find_one(
        self, prefix: str = ..., user_role: str = ..., account_status: str = ...
    ) -> User | None: ...

    @overload
    def find_one(self, *args, **kwargs) -> User | None: ...

    def find_one(self, *args, **kwargs) -> User | None:
        params = dict(*args, **kwargs)
        paginator = Paginator(self.session, self.url, params=params)
        pages = paginator.fetch_pages()
        results = (result for page in pages for result in page.results)
        users = (
            User(
                config=self.config,
                session=self.session,
                **result,
            )
            for result in results
        )
        return next(users, None)

    def get(self, id: str) -> User:
        url = urls.append_path(self.config.url, f"v1/users/{id}")
        response = self.session.get(url)
        return User(
            config=self.config,
            session=self.session,
            **response.json(),
        )

    def create(self) -> User:
        raise NotImplementedError()

    def update(self) -> User:
        raise NotImplementedError()

    def delete(self) -> None:
        raise NotImplementedError()

    def count(self) -> int:
        response: requests.Response = self.session.get(
            self.url, params={"page_size": 1}
        )
        result: dict = response.json()
        return result["total"]



import pyjson5 as json

from pathlib import Path


def load_mock(path: str) -> dict:
    """
    Load mock data from a file.

    Reads a JSON or JSONC (JSON with Comments) file and returns the parsed data.

    It's primarily used for loading mock data for tests.

    The file names for mock data should match the query path that they represent.

    Parameters
    ----------
    path : str
        The relative path to the JSONC file.

    Returns
    -------
    dict
        The parsed data from the JSONC file.

    Examples
    --------
    >>> data = load_mock("v1/example.json")
    >>> data = load_mock("v1/example.jsonc")
    """
    return json.loads((Path(__file__).parent / "__api__" / path).read_text())



# -*- coding: utf-8 -*-
# mypy: ignore-errors
import os

from posit.connect.external.databricks import viewer_credentials_provider
from databricks import sql
from flask import Flask, request

DATABRICKS_HOST = os.getenv("DATABRICKS_HOST")
DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
SQL_HTTP_PATH = os.getenv("DATABRICKS_PATH")

rows = None
app = Flask(__name__)


@app.route("/")
def usage():
    return "<p>Try: <pre>GET /fares<pre></p>"


@app.route("/fares")
def get_fares():
    """
    Flask example API that returns the first few rows from
    a table hosted in Databricks.
    """
    global rows

    session_token = request.headers.get("Posit-Connect-User-Session-Token")
    credentials_provider = viewer_credentials_provider(user_session_token=session_token)

    if rows is None:
        query = "SELECT * FROM samples.nyctaxi.trips LIMIT 10;"

        with sql.connect(
            server_hostname=DATABRICKS_HOST,
            http_path=SQL_HTTP_PATH,
            auth_type="databricks-oauth",
            credentials_provider=credentials_provider,
        ) as connection:
            with connection.cursor() as cursor:
                cursor.execute(query)
                rows = cursor.fetchall()

    return [row.asDict() for row in rows]


if __name__ == "__main__":
    app.run(debug=True)



import pytest

from unittest.mock import patch

from posit.connect.config import Config, _get_api_key, _get_url


@patch.dict("os.environ", {"CONNECT_API_KEY": "foobar"})
def test_get_api_key():
    api_key = _get_api_key()
    assert api_key == "foobar"


@patch.dict("os.environ", {"CONNECT_API_KEY": ""})
def test_get_api_key_empty():
    with pytest.raises(ValueError):
        _get_api_key()


@patch.dict("os.environ", clear=True)
def test_get_api_key_miss():
    with pytest.raises(ValueError):
        _get_api_key()


@patch.dict("os.environ", {"CONNECT_SERVER": "http://foo.bar"})
def test_get_url():
    url = _get_url()
    assert url == "http://foo.bar"


@patch.dict("os.environ", {"CONNECT_SERVER": ""})
def test_get_url_empty():
    with pytest.raises(ValueError):
        _get_url()


@patch.dict("os.environ", clear=True)
def test_get_url_miss():
    with pytest.raises(ValueError):
        _get_url()


def test_init():
    api_key = "foobar"
    url = "http://foo.bar/__api__"
    config = Config(api_key=api_key, url=url)
    assert config.api_key == api_key
    assert config.url == url



from __future__ import annotations

import posixpath

from urllib.parse import urlsplit, urlunsplit


def server_to_api_url(url: str) -> str:
    """
    Fixes the given URL by appending '__api__' if it doesn't already end with it.

    Args:
        url (str): The URL to fix.

    Returns:
        str: The fixed URL.
    """
    url = url.rstrip("/")
    if not url.endswith("__api__"):
        return append_path(url, "__api__")
    return url


def validate(url: str) -> None:
    """
    Check if the given URL is valid.

    Args:
        url (str): The URL to be validated.

    Returns:
        bool: True if the URL is valid, False otherwise.

    Raises:
        ValueError: If the URL is missing a scheme or is not absolute.
    """
    split = urlsplit(url, allow_fragments=False)
    if not split.scheme:
        raise ValueError(
            f"url must specify a scheme (e.g., http://example.com/__api__): {url}"
        )

    if not split.netloc:
        raise ValueError(
            f"url must be absolute (e.g., http://example.com/__api__): {url}"
        )


def append_path(url: str, path: str) -> str:
    """
    Appends a path to a URL.

    Args:
        url (str): The original URL.
        path (str): The path to append.

    Returns:
        str: The modified URL with the appended path.
    """
    # See https://docs.python.org/3/library/urllib.parse.html#urllib.parse.urlsplit
    split = urlsplit(url, allow_fragments=False)
    # Removes leading '/' from path to avoid double slashes.
    path = path.lstrip("/")
    # Prepend the path with the existing URL path.
    path = posixpath.join(split.path, path)
    # See https://docs.python.org/3/library/urllib.parse.html#urllib.parse.urlunsplit
    return urlunsplit((split.scheme, split.netloc, path, split.query, None))



from requests import PreparedRequest
from requests.auth import AuthBase

from .config import Config


class Auth(AuthBase):
    def __init__(self, config: Config) -> None:
        self._config = config

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        r.headers["Authorization"] = f"Key {self._config.api_key}"
        return r



import io

import pytest

from requests import HTTPError, Response
from unittest.mock import Mock, patch

from posit.connect.errors import ClientError
from posit.connect.hooks import handle_errors


def test_success():
    response = Mock()
    response.status_code = 200
    assert handle_errors(response) == response


def test_client_error():
    response = Mock()
    response.status_code = 400
    response.json = Mock(return_value={"code": 0, "error": "foobar"})
    with pytest.raises(ClientError):
        handle_errors(response)


@patch("posit.connect.hooks.JSONDecodeError")
def test_client_error_without_payload(JSONDecodeError):
    response = Mock()
    response.status_code = 404
    response.json = Mock(side_effect=JSONDecodeError())
    response.raise_for_status = Mock(side_effect=Exception())
    with pytest.raises(Exception):
        handle_errors(response)


def test_200():
    response = Response()
    response.status_code = 200
    assert handle_errors(response) == response


def test_response_client_error_with_plaintext_payload():
    response = Response()
    response.status_code = 404
    response.raw = io.BytesIO(b"Plain text 404 Not Found")
    with pytest.raises(HTTPError):
        handle_errors(response)


def test_response_client_error_with_json_payload():
    response = Response()
    response.status_code = 400
    response.raw = io.BytesIO(b'{"code":0,"error":"foobar"}')
    with pytest.raises(
        ClientError, match=r"foobar \(Error Code: 0, HTTP Status: 400 Bad Request\)"
    ):
        handle_errors(response)


def test_response_client_error_without_payload():
    response = Response()
    response.status_code = 404
    response.raw = io.BytesIO(b"Plain text 404 Not Found")
    with pytest.raises(HTTPError):
        handle_errors(response)
